class HttpMethod(object):
    HEAD = 'HEAD'
    GET = 'GET'
    POST = 'POST'
    PUT = 'PUT'
    DELETE = 'DELETE'


class ResponseCodes(object):
    OK = 200
    CREATED = 201
    ACCEPTED = 202
    BAD_REQUEST = 400
    NOT_FOUND = 404


class HostParsing(object):
    HTTP = 'http'
    HTTPS = 'https'
    USE_SSL = 'use_ssl'
    URL_PREFIX = 'url_prefix'
    SCHEME = 'scheme'
    HTTP_AUTH = 'http_auth'
    PORT = 'port'
    HOST = 'host'


class EsMethods(object):
    MULTIPLE_SEARCH = '_msearch'
    BULK = '_bulk'
    COUNT = '_count'
    EXPLAIN = '_explain'
    MULTIPLE_GET = '_mget'
    UPDATE = '_update'
    SOURCE = '_source'
    SEARCH = '_search'
    SCROLL = 'scroll'


class EsConst(object):
    TOTAL = 'total'
    FAILED = 'failed'
    SHARDS = '_shards'
    ALL_VALUES = '_all'
    FIELDS = 'fields'
    SCROLL_ID = 'scroll_id'
    HITS = 'hits'
    FOUND = 'found'


class EsBulk(object):
    OP_TYPE = '_op_type'
    INDEX = 'index'
    CREATE = 'create'
    DELETE = 'delete'
    UPDATE = 'update'


class EsDocProperties(object):
    INDEX = '_index'
    TYPE = '_type'
    PARENT = '_parent'
    PERCOLATE = '_percolate'
    ROUTING = '_routing'
    TIMESTAMP = '_timestamp'
    TTL = '_ttl'
    VERSION = '_version'
    VERSION_TYPE = '_version_type'
    ID = '_id'
    RETRY_ON_CONFLICT = '_retry_on_conflict'
    SOURCE = '_source'
    SCROLL_ID = '_scroll_id'

# parts of URL to be omitted
NULL_VALUES = (None, '', b'', [], ())



from twisted.internet.defer import succeed, inlineCallbacks, returnValue

from twistes.consts import EsDocProperties
from twistes.utilities import EsUtils


class Scroller(object):
    """
    Handle scrolling through scan and scroll API.

    Usage: after the creation of the Scroller,
    Iterate over it and yield each item to get the results.
    Example:
        scroller = es.scan(...)
        for item in scroller:
            results = yield item
            for hit in results:
                ...
    """

    def __init__(self, es, results, scroll, size):
        self._first_results = results
        self._scroll_id = results.get(EsDocProperties.SCROLL_ID, None)
        self._scroll = scroll
        self._size = size
        self._es = es

    def __iter__(self):
        return self

    def next(self):
        """Fetch next page from scroll API."""
        d = None
        if self._first_results:
            d = succeed(EsUtils.extract_hits(self._first_results))
            self._first_results = None
        elif self._scroll_id:
            d = self._scroll_next_results()
        else:
            raise StopIteration()
        return d

    def __next__(self):
        return self.next()

    @inlineCallbacks
    def _scroll_next_results(self):
        results = yield self._es.scroll(str(self._scroll_id), scroll=self._scroll)
        hits = EsUtils.extract_hits(results)

        # No more results
        if len(hits) < self._size:
            self._scroll_id = None
        else:
            self._scroll_id = results.get(EsDocProperties.SCROLL_ID, None)

        returnValue(hits)



from twistes.compatability import quote, urlencode, string_types, urlparse

from twistes.consts import NULL_VALUES, HostParsing


class EsParser(object):
    SSL_DEFAULT_PORT = 443

    @staticmethod
    def parse_host(hosts):
        """
        Parsing the hosts parameter,
        * currently support only one host
        :param hosts: the hosts json to parse
        :return: the full host and the authentication if exists
        """
        hosts = EsParser._normalize_hosts(hosts)
        host = hosts[0]

        host_name = host[HostParsing.HOST]
        host_port = host[HostParsing.PORT]

        auth = None
        if HostParsing.HTTP_AUTH in host:
            http_auth = host[HostParsing.HTTP_AUTH]
            user_pass = http_auth.split(':')
            auth = (user_pass[0], user_pass[1])

        full_host = "{host}:{port}".format(host=host_name, port=host_port)
        if not host_name.startswith((HostParsing.HTTP + ':', HostParsing.HTTPS + ':')):
            scheme = HostParsing.HTTPS if host.get(HostParsing.USE_SSL) else HostParsing.HTTP
            full_host = "{scheme}://{full_host}".format(full_host=full_host, scheme=scheme)

        return full_host, auth

    @staticmethod
    def _normalize_hosts(hosts):
        if hosts is None:
            hosts = [{}]

        if isinstance(hosts, string_types):
            hosts = [hosts]

        return list(map(EsParser._normalize_host, hosts))

    @staticmethod
    def _normalize_host(host):
        if isinstance(host, string_types):
            return EsParser._parse_string_host(host)
        else:
            return EsParser._update_ssl_params(host)

    @staticmethod
    def _update_ssl_params(host):
        """
        Update the host ssl params (port or scheme) if needed.
        :param host:
        :return:
        """
        if host[HostParsing.HOST] \
                and EsParser._is_secure_connection_type(host):
            host[HostParsing.PORT] = EsParser.SSL_DEFAULT_PORT
            host[HostParsing.USE_SSL] = True
            parsed_url = urlparse(EsParser._fix_host_prefix(host[HostParsing.HOST]))
            host[HostParsing.HOST] = parsed_url.hostname
            host[HostParsing.SCHEME] = HostParsing.HTTPS
        return host

    @staticmethod
    def _parse_string_host(host_str):
        """
        Parse host string into a dictionary host
        :param host_str:
        :return:
        """
        host_str = EsParser._fix_host_prefix(host_str)
        parsed_url = urlparse(host_str)
        host = {HostParsing.HOST: parsed_url.hostname}
        if parsed_url.port:
            host[HostParsing.PORT] = parsed_url.port
        if parsed_url.scheme == HostParsing.HTTPS:
            host[HostParsing.PORT] = parsed_url.port or EsParser.SSL_DEFAULT_PORT
            host[HostParsing.USE_SSL] = True
            host[HostParsing.SCHEME] = HostParsing.HTTPS
        elif parsed_url.scheme:
            host[HostParsing.SCHEME] = parsed_url.scheme
        if parsed_url.username or parsed_url.password:
            host[HostParsing.HTTP_AUTH] = '%s:%s' % (parsed_url.username, parsed_url.password)
        if parsed_url.path and parsed_url.path != '/':
            host[HostParsing.URL_PREFIX] = parsed_url.path
        return host

    @staticmethod
    def _fix_host_prefix(host):
        if '://' not in host:
            host = "//%s" % host
        return host

    @staticmethod
    def _is_secure_connection_type(host):
        return (HostParsing.HTTPS in host[HostParsing.HOST] and not host.get(HostParsing.PORT)) \
                or (HostParsing.HTTPS not in host[HostParsing.HOST] and host.get(HostParsing.PORT, 0) == 443)

    @staticmethod
    def make_path(*sub_paths):
        """
        Create a path from a list of sub paths.
        :param sub_paths: a list of sub paths
        :return:
        """
        queued_params = [quote(str(c), '') for c in sub_paths if c not in NULL_VALUES]
        queued_params.insert(0, '')
        return '/'.join(queued_params)

    @staticmethod
    def prepare_url(hostname, path, params=None):
        """
        Prepare Elasticsearch request url.
        :param hostname: host name
        :param path: request path
        :param params: optional url params
        :return:
        """
        url = hostname + path

        if params:
            url = url + '?' + urlencode(params)

        if not url.startswith(('http:', 'https:')):
            url = "http://" + url

        return url.encode('utf-8')

    @staticmethod
    def is_not_empty_params(*kwargs):
        for param in kwargs:
            if param in NULL_VALUES:
                raise ValueError("Empty value passed for a required argument {empty_param}.".format(empty_param=param))



import json
from operator import methodcaller

from twisted.internet.defer import inlineCallbacks, returnValue

from twistes.compatability import string_types
from twistes.consts import EsBulk, EsDocProperties
from twistes.exceptions import BulkIndexError, ConnectionTimeout


class ActionParser(object):
    ES_OPERATIONS_PARAMS = (
        EsDocProperties.INDEX, EsDocProperties.PARENT, EsDocProperties.PERCOLATE, EsDocProperties.ROUTING,
        EsDocProperties.TIMESTAMP, EsDocProperties.TTL, EsDocProperties.TYPE, EsDocProperties.VERSION,
        EsDocProperties.VERSION_TYPE, EsDocProperties.ID, EsDocProperties.RETRY_ON_CONFLICT)

    @staticmethod
    def expand_action(data):
        """
        From one document or action definition passed in by the user extract the
        action/data lines needed for elasticsearch's
        :meth:`~elasticsearch.Elasticsearch.bulk` api.
        :return es format to bulk doc
        """
        # when given a string, assume user wants to index raw json
        if isinstance(data, string_types):
            return '{"index": {}}', data

        # make sure we don't alter the action
        data = data.copy()
        op_type = data.pop(EsBulk.OP_TYPE, EsBulk.INDEX)

        action = ActionParser._get_relevant_action_params(data, op_type)

        # no data payload for delete
        if op_type == EsBulk.DELETE:
            return action, None

        return action, data.get(EsDocProperties.SOURCE, data)

    @staticmethod
    def _get_relevant_action_params(data, op_type):
        action = {op_type: {}}
        for key in ActionParser.ES_OPERATIONS_PARAMS:
            if key in data:
                action[op_type][key] = data.pop(key)

        return action


class BulkUtility(object):

    def __init__(self, es):
        self.client = es

    @inlineCallbacks
    def bulk(self, actions, stats_only=False, **kwargs):
        """
        Helper for the :meth:`~elasticsearch.Elasticsearch.bulk` api that provides
        a more human friendly interface - it consumes an iterator of actions and
        sends them to elasticsearch in chunks. It returns a tuple with summary
        information - number of successfully executed actions and either list of
        errors or number of errors if `stats_only` is set to `True`.
        See :func:`~elasticsearch.helpers.streaming_bulk` for more accepted
        parameters
        :arg actions: iterator containing the actions
        :arg stats_only: if `True` only report number of successful/failed
            operations instead of just number of successful and a list of error responses
        Any additional keyword arguments will be passed to
        :func:`~elasticsearch.helpers.streaming_bulk` which is used to execute
        the operation.
        """
        success, failed = 0, 0

        # list of errors to be collected is not stats_only
        errors = []

        for deferred_bulk in self.streaming_bulk(actions, **kwargs):
            bulk_results = yield deferred_bulk
            for ok, item in bulk_results:
                # go through request-response pairs and detect failures
                if not ok:
                    if not stats_only:
                        errors.append(item)
                    failed += 1
                else:
                    success += 1
        summarized_results = success, failed if stats_only else errors
        returnValue(summarized_results)

    def streaming_bulk(self, actions, chunk_size=500, max_chunk_bytes=100 * 1024 * 1024,
                       raise_on_error=True, expand_action_callback=ActionParser.expand_action,
                       raise_on_exception=True, **kwargs):
        """
        Streaming bulk consumes actions from the iterable passed in and return the results of all bulk data
        :func:`~elasticsearch.helpers.bulk` which is a wrapper around streaming
        bulk that returns summary information about the bulk operation once the
        entire input is consumed and sent.
        :arg actions: iterable containing the actions to be executed
        :arg chunk_size: number of docs in one chunk sent to es (default: 500)
        :arg max_chunk_bytes: the maximum size of the request in bytes (default: 100MB)
        :arg raise_on_error: raise ``BulkIndexError`` containing errors (as `.errors`)
            from the execution of the last chunk when some occur. By default we raise.
        :arg raise_on_exception: if ``False`` then don't propagate exceptions from
            call to ``bulk`` and just report the items that failed as failed.
        :arg expand_action_callback: callback executed on each action passed in,
            should return a tuple containing the action line and the data line
            (`None` if data line should be omitted).
        """
        actions = list(map(expand_action_callback, actions))

        for bulk_actions in self._chunk_actions(actions, chunk_size, max_chunk_bytes):
            yield self._process_bulk_chunk(bulk_actions, raise_on_exception, raise_on_error, **kwargs)

    @staticmethod
    def _chunk_actions(actions, chunk_size, max_chunk_bytes):
        """
        Split actions into chunks by number or size, serialize them into strings in
        the process.
        """
        bulk_actions = []
        size, action_count = 0, 0
        for action, data in actions:
            action = json.dumps(action)
            cur_size = len(action) + 1

            if data is not None:
                data = json.dumps(data)
                cur_size += len(data) + 1

            # full chunk, send it and start a new one
            if bulk_actions and (size + cur_size > max_chunk_bytes or action_count == chunk_size):
                yield bulk_actions
                bulk_actions = []
                size, action_count = 0, 0

            bulk_actions.append(action)
            if data is not None:
                bulk_actions.append(data)
            size += cur_size
            action_count += 1

        if bulk_actions:
            yield bulk_actions

    @inlineCallbacks
    def _process_bulk_chunk(self, bulk_actions, raise_on_exception=True, raise_on_error=True, **kwargs):
        """
        Send a bulk request to elasticsearch and process the output.
        """
        # if raise on error is set, we need to collect errors per chunk before
        # raising them

        resp = None
        try:
            # send the actual request
            actions = "{}\n".format('\n'.join(bulk_actions))
            resp = yield self.client.bulk(actions, **kwargs)
        except ConnectionTimeout as e:
            # default behavior - just propagate exception
            if raise_on_exception:
                raise

            self._handle_transport_error(bulk_actions, e, raise_on_error)
            returnValue([])

        # go through request-response pairs and detect failures
        errors = []
        results = []
        for op_type, item in map(methodcaller('popitem'), resp['items']):
            ok = 200 <= item.get('status', 500) < 300
            if not ok and raise_on_error:
                errors.append({op_type: item})

            if ok or not errors:
                # if we are not just recording all errors to be able to raise
                # them all at once, yield items individually
                results.append((ok, {op_type: item}))

        if errors:
            msg_fmt = '{num} document(s) failed to index.'
            raise BulkIndexError(msg_fmt.format(num=len(errors)), errors)
        else:
            returnValue(results)

    @staticmethod
    def _handle_transport_error(bulk_actions, e, raise_on_error):
        # if we are not propagating, mark all actions in current chunk as
        # failed
        exc_errors = []
        # deserialize the data back, this is expensive but only run on
        # errors if raise_on_exception is false, so shouldn't be a real
        # issue
        bulk_data = iter(map(json.loads, bulk_actions))
        while True:
            try:
                # collect all the information about failed actions
                action = next(bulk_data)
                op_type, action = action.popitem()
                # TODO:: add the status code after we add it to the exception
                info = {"error": str(e), "exception": e}
                if op_type != 'delete':
                    info['data'] = next(bulk_data)
                info.update(action)
                exc_errors.append({op_type: info})
            except StopIteration:
                break

        # emulate standard behavior for failed actions
        if raise_on_error:
            msg_fmt = '{num} document(s) failed to index.'
            raise BulkIndexError(msg_fmt.format(num=len(exc_errors)),
                                 exc_errors)



class ElasticsearchException(Exception):
    """
    Base class for all exceptions raised by this package's operations (doesn't
    apply to :class:`~elasticsearch.ImproperlyConfigured`).
    """


class ImproperlyConfigured(Exception):
    """
    Exception raised when the config passed to the client is inconsistent or invalid.
    """


class ScanError(ElasticsearchException):
    """An Error eccure during the scan and scroll request"""


class BulkIndexError(ElasticsearchException):

    @property
    def errors(self):
        """ List of errors from execution of the last chunk. """
        return self.args[1]


class SerializationError(ElasticsearchException):
    """
    Data passed in failed to serialize properly in the Serializer being used.
    """


class TransportError(ElasticsearchException):
    """
    Exception raised when ES returns a non-OK (>=400) HTTP status code.
    Or when an actual connection error happens;  in that case the status_code
    will be set to 'N/A'.
    """

    def __init__(self, error, status_code=None, info=None):
        super(TransportError, self).__init__(error)
        self.error = error
        self.status_code = status_code or 'N/A'
        self.info = info or {}


class ConnectionError(TransportError):
    """
    Error raised when there was an exception while talking to ES.
    Original exception from the underlying Connection implementation
    is available as .info.
    """

    def __init__(self, error):
        super(ConnectionError, self).__init__(error=error)


class ConnectionTimeout(ConnectionError):
    """
    A network timeout. DOesn't cause a node retry by default.
    """


class SSLError(ConnectionError):
    """
    Error raised when encountering SSL errors.
    """


class NotFoundError(TransportError):
    """
    Exception representing a 404 status code.
    """

    def __init__(self, info):
        super(NotFoundError, self).__init__(error="not found",
                                            info=info,
                                            status_code=404)


class ConflictError(TransportError):
    """
    Exception representing a 409 status code.
    """

    def __init__(self, info):
        super(ConflictError, self).__init__(error="conflict",
                                            info=info,
                                            status_code=409)


class RequestError(TransportError):
    """
    Exception representing a 400 status code.
    """

    def __init__(self, info):
        super(RequestError, self).__init__(error="bad request",
                                           info=info,
                                           status_code=400)



import treq
import json
from twisted.internet.defer import inlineCallbacks, returnValue, CancelledError
from twisted.internet.error import ConnectingCancelledError
from twisted.web._newclient import ResponseNeverReceived
from twistes.compatability import string_types
from twistes.exceptions import (NotFoundError,
                                ConnectionTimeout,
                                RequestError,
                                ElasticsearchException)
from twistes.scroller import Scroller
from twistes.consts import HttpMethod, EsMethods, EsConst, NULL_VALUES
from twistes.parser import EsParser
from twistes.consts import ResponseCodes
from twistes.bulk_utils import BulkUtility


class Elasticsearch(object):
    """
    Elastic search asynchronous http client implemented with treq and twisted
    """

    def __init__(self, hosts, timeout=10, async_http_client=treq):
        self._es_parser = EsParser()
        self._hostname, self._auth = self._es_parser.parse_host(hosts)
        self._timeout = timeout
        self._async_http_client = async_http_client
        self.bulk_utils = BulkUtility(self)

    @inlineCallbacks
    def info(self, **query_params):
        """
        Get basic information about the cluster
        `<http://www.elastic.co/guide/>`_
        """
        yield self._perform_request(HttpMethod.GET, '/', params=query_params)

    @inlineCallbacks
    def get(self, index, id, fields=None, doc_type=EsConst.ALL_VALUES, **query_params):
        """
        Retrieve specific record by id
        `<http://www.elastic.co/guide/en/elasticsearch/reference/current/docs-get.html>`_
        :param index: the index name to query
        :param id: the id of the record
        :param fields: the fields you what to fetch from the record (str separated by comma's)
        :param doc_type: the doc type to search in
        :param query_params: params
        :return:
        """
        if fields:
            query_params[EsConst.FIELDS] = fields

        path = self._es_parser.make_path(index, doc_type, id)
        result = yield self._perform_request(HttpMethod.GET, path, params=query_params)
        returnValue(result)

    @inlineCallbacks
    def exists(self, index, doc_type, id, **query_params):
        """
        Check if the doc exist in the elastic search
        `<http://www.elastic.co/guide/en/elasticsearch/reference/current/docs-get.html>`_
        :param index: the index name
        :param doc_type: the document type
        :param id: the id of the doc type
        :arg parent: The ID of the parent document
        :arg preference: Specify the node or shard the operation should be
            performed on (default: random)
        :arg realtime: Specify whether to perform the operation in realtime or
            search mode
        :arg refresh: Refresh the shard containing the document before
            performing the operation
        :arg routing: Specific routing value
        """
        self._es_parser.is_not_empty_params(index, doc_type, id)
        path = self._es_parser.make_path(index, doc_type, id)
        result = yield self._perform_request(HttpMethod.HEAD,
                                             path,
                                             params=query_params)
        returnValue(result)

    @inlineCallbacks
    def get_source(self, index, doc_type, id, **query_params):
        """
        Get the _source of the document
        `<http://www.elastic.co/guide/en/elasticsearch/reference/current/docs-get.html>`_
        :param index: the index name
        :param doc_type: the document type
        :param id: the id of the doc type
        :arg _source: True or false to return the _source field or not, or a
            list of fields to return
        :arg _source_exclude: A list of fields to exclude from the returned
            _source field
        :arg _source_include: A list of fields to extract and return from the
            _source field
        :arg parent: The ID of the parent document
        :arg preference: Specify the node or shard the operation should be
            performed on (default: random)
        :arg realtime: Specify whether to perform the operation in realtime or
            search mode
        :arg refresh: Refresh the shard containing the document before
            performing the operation
        :arg routing: Specific routing value
        :arg version: Explicit version number for concurrency control
        :arg version_type: Specific version type, valid choices are: 'internal',
            'external', 'external_gte', 'force'
        """
        self._es_parser.is_not_empty_params(index, doc_type, id)
        path = self._es_parser.make_path(index, doc_type, id, EsMethods.SOURCE)
        result = yield self._perform_request(HttpMethod.GET,
                                             path,
                                             params=query_params)
        returnValue(result)

    @inlineCallbacks
    def mget(self, body, index=None, doc_type=None, **query_params):
        """
        Get multiple document from the same index and doc_type (optionally) by ids
        `<http://www.elastic.co/guide/en/elasticsearch/reference/current/docs-multi-get.html>`

        :param body: list of docs with or without the index and type parameters
        or list of ids
        :param index: the name of the index
        :param doc_type: the document type
        :arg _source: True or false to return the _source field or not, or a
            list of fields to return
        :arg _source_exclude: A list of fields to exclude from the returned
            _source field
        :arg _source_include: A list of fields to extract and return from the
            _source field
        :arg fields: A comma-separated list of fields to return in the response
        :arg preference: Specify the node or shard the operation should be
            performed on (default: random)
        :arg realtime: Specify whether to perform the operation in realtime or
            search mode
        :arg refresh: Refresh the shard containing the document before
            performing the operation
        """
        self._es_parser.is_not_empty_params(body)

        path = self._es_parser.make_path(index,
                                         doc_type,
                                         EsMethods.MULTIPLE_GET)

        result = yield self._perform_request(HttpMethod.GET,
                                             path,
                                             body=body,
                                             params=query_params)
        returnValue(result)

    @inlineCallbacks
    def update(self, index, doc_type, id, body=None, **query_params):
        """
        Update a document with the body param or list of ids
        `<http://www.elastic.co/guide/en/elasticsearch/reference/current/docs-update.html>`_

        :param index: the name of the index
        :param doc_type: the document type
        :param id: the id of the doc type
        :param body: the data to update in the index
        :arg consistency: Explicit write consistency setting for the operation,
            valid choices are: 'one', 'quorum', 'all'
        :arg fields: A comma-separated list of fields to return in the response
        :arg lang: The script language (default: groovy)
        :arg parent: ID of the parent document. Is is only used for routing and
            when for the upsert request
        :arg refresh: Refresh the index after performing the operation
        :arg retry_on_conflict: Specify how many times should the operation be
            retried when a conflict occurs (default: 0)
        :arg routing: Specific routing value
        :arg script: The URL-encoded script definition (instead of using request
            body)
        :arg script_id: The id of a stored script
        :arg scripted_upsert: True if the script referenced in script or
            script_id should be called to perform inserts - defaults to false
        :arg timeout: Explicit operation timeout
        :arg timestamp: Explicit timestamp for the document
        :arg ttl: Expiration time for the document
        :arg version: Explicit version number for concurrency control
        :arg version_type: Specific version type, valid choices are: 'internal',
            'force'
        """
        self._es_parser.is_not_empty_params(index, doc_type, id)
        path = self._es_parser.make_path(index, doc_type, id, EsMethods.UPDATE)
        result = yield self._perform_request(HttpMethod.POST, path, body=body, params=query_params)
        returnValue(result)

    @inlineCallbacks
    def search(self, index=None, doc_type=None, body=None, **query_params):
        """
        Make a search query on the elastic search
        `<http://www.elastic.co/guide/en/elasticsearch/reference/current/search-search.html>`_

        :param index: the index name to query
        :param doc_type: he doc type to search in
        :param body: the query
        :param query_params: params
        :arg _source: True or false to return the _source field or not, or a
            list of fields to return
        :arg _source_exclude: A list of fields to exclude from the returned
            _source field
        :arg _source_include: A list of fields to extract and return from the
            _source field
        :arg allow_no_indices: Whether to ignore if a wildcard indices
            expression resolves into no concrete indices. (This includes `_all`
            string or when no indices have been specified)
        :arg analyze_wildcard: Specify whether wildcard and prefix queries
            should be analyzed (default: false)
        :arg analyzer: The analyzer to use for the query string
        :arg default_operator: The default operator for query string query (AND
            or OR), default 'OR', valid choices are: 'AND', 'OR'
        :arg df: The field to use as default where no field prefix is given in
            the query string
        :arg expand_wildcards: Whether to expand wildcard expression to concrete
            indices that are open, closed or both., default 'open', valid
            choices are: 'open', 'closed', 'none', 'all'
        :arg explain: Specify whether to return detailed information about score
            computation as part of a hit
        :arg fielddata_fields: A comma-separated list of fields to return as the
            field data representation of a field for each hit
        :arg fields: A comma-separated list of fields to return as part of a hit
        :arg from\_: Starting offset (default: 0)
        :arg ignore_unavailable: Whether specified concrete indices should be
            ignored when unavailable (missing or closed)
        :arg lenient: Specify whether format-based query failures (such as
            providing text to a numeric field) should be ignored
        :arg lowercase_expanded_terms: Specify whether query terms should be
            lowercased
        :arg preference: Specify the node or shard the operation should be
            performed on (default: random)
        :arg q: Query in the Lucene query string syntax
        :arg request_cache: Specify if request cache should be used for this
            request or not, defaults to index level setting
        :arg routing: A comma-separated list of specific routing values
        :arg scroll: Specify how long a consistent view of the index should be
            maintained for scrolled search
        :arg search_type: Search operation type, valid choices are:
            'query_then_fetch', 'dfs_query_then_fetch'
        :arg size: Number of hits to return (default: 10)
        :arg sort: A comma-separated list of <field>:<direction> pairs
        :arg stats: Specific 'tag' of the request for logging and statistical
            purposes
        :arg suggest_field: Specify which field to use for suggestions
        :arg suggest_mode: Specify suggest mode, default 'missing', valid
            choices are: 'missing', 'popular', 'always'
        :arg suggest_size: How many suggestions to return in response
        :arg suggest_text: The source text for which the suggestions should be
            returned
        :arg terminate_after: The maximum number of documents to collect for
            each shard, upon reaching which the query execution will terminate
            early.
        :arg timeout: Explicit operation timeout
        :arg track_scores: Whether to calculate and return scores even if they
            are not used for sorting
        :arg version: Specify whether to return document version as part of a
            hit
        """
        path = self._es_parser.make_path(index, doc_type, EsMethods.SEARCH)
        result = yield self._perform_request(HttpMethod.POST, path, body=body, params=query_params)
        returnValue(result)

    @inlineCallbacks
    def explain(self, index, doc_type, id, body=None, **query_params):
        """
        The explain api computes a score explanation for a query and a specific
        document. This can give useful feedback whether a document matches or
        didn't match a specific query.
        `<http://www.elastic.co/guide/en/elasticsearch/reference/current/search-explain.html>`_
        :param index: The name of the index
        :param doc_type: The type of the document
        :param id: The document ID
        :param body: The query definition using the Query DSL
        :arg _source: True or false to return the _source field or not, or a
            list of fields to return
        :arg _source_exclude: A list of fields to exclude from the returned
            _source field
        :arg _source_include: A list of fields to extract and return from the
            _source field
        :arg analyze_wildcard: Specify whether wildcards and prefix queries in
            the query string query should be analyzed (default: false)
        :arg analyzer: The analyzer for the query string query
        :arg default_operator: The default operator for query string query (AND
            or OR), default 'OR', valid choices are: 'AND', 'OR'
        :arg df: The default field for query string query (default: _all)
        :arg fields: A comma-separated list of fields to return in the response
        :arg lenient: Specify whether format-based query failures (such as
            providing text to a numeric field) should be ignored
        :arg lowercase_expanded_terms: Specify whether query terms should be
            lowercased
        :arg parent: The ID of the parent document
        :arg preference: Specify the node or shard the operation should be
            performed on (default: random)
        :arg q: Query in the Lucene query string syntax
        :arg routing: Specific routing value
        """
        self._es_parser.is_not_empty_params(index, doc_type, id)

        path = self._es_parser.make_path(index,
                                         doc_type,
                                         id,
                                         EsMethods.EXPLAIN)

        result = yield self._perform_request(HttpMethod.GET,
                                             path,
                                             body,
                                             params=query_params)
        returnValue(result)

    @inlineCallbacks
    def delete(self, index, doc_type, id, **query_params):
        """
        Delete specific record by id
        `<http://www.elastic.co/guide/en/elasticsearch/reference/current/docs-delete.html>`_
        :param index: the index name to delete from
        :param doc_type: the doc type to delete from
        :param id: the id of the record
        :param query_params: params
        :return:
        """
        path = self._es_parser.make_path(index, doc_type, id)
        result = yield self._perform_request(HttpMethod.DELETE, path, params=query_params)
        returnValue(result)

    @inlineCallbacks
    def index(self, index, doc_type, body, id=None, **query_params):
        """
        Adds or updates a typed JSON document in a specific index, making it searchable.
        `<http://www.elastic.co/guide/en/elasticsearch/reference/current/docs-index_.html>`_

        :param index: The name of the index
        :param doc_type: The type of the document
        :param body: The document
        :param id: Document ID

        :arg consistency: Explicit write consistency setting for the operation,
            valid choices are: 'one', 'quorum', 'all'
        :arg op_type: Explicit operation type, default 'index', valid choices
            are: 'index', 'create'
        :arg parent: ID of the parent document
        :arg refresh: Refresh the index after performing the operation
        :arg routing: Specific routing value
        :arg timeout: Explicit operation timeout
        :arg timestamp: Explicit timestamp for the document
        :arg ttl: Expiration time for the document
        :arg version: Explicit version number for concurrency control
        :arg version_type: Specific version type, valid choices are: 'internal',
            'external', 'external_gte', 'force'
        """
        self._es_parser.is_not_empty_params(index, doc_type, body)

        method = HttpMethod.POST if id in NULL_VALUES else HttpMethod.PUT
        path = self._es_parser.make_path(index, doc_type, id)
        result = yield self._perform_request(method, path, body, params=query_params)
        returnValue(result)

    @inlineCallbacks
    def create(self, index, doc_type, body, id=None, **query_params):
        """
        Adds a typed JSON document in a specific index, making it searchable.
        Behind the scenes this method calls index(..., op_type='create')
        `<http://www.elastic.co/guide/en/elasticsearch/reference/current/docs-index_.html>`_
        :param index: The name of the index
        :param doc_type: The type of the document
        :param body: The document
        :param id: Document ID
        :arg consistency: Explicit write consistency setting for the operation,
            valid choices are: 'one', 'quorum', 'all'
        :arg op_type: Explicit operation type, default 'index', valid choices
            are: 'index', 'create'
        :arg parent: ID of the parent document
        :arg refresh: Refresh the index after performing the operation
        :arg routing: Specific routing value
        :arg timeout: Explicit operation timeout
        :arg timestamp: Explicit timestamp for the document
        :arg ttl: Expiration time for the document
        :arg version: Explicit version number for concurrency control
        :arg version_type: Specific version type, valid choices are: 'internal',
            'external', 'external_gte', 'force'
        """
        query_params['op_type'] = 'create'
        result = yield self.index(index, doc_type, body, id=id, params=query_params)
        returnValue(result)

    @inlineCallbacks
    def scroll(self, scroll_id=None, body=None, **query_params):
        """
        Scroll a search request created by specifying the scroll parameter.
        `<http://www.elastic.co/guide/en/elasticsearch/reference/current/search-request-scroll.html>`_

        :param scroll_id: The scroll ID
        :param body: The scroll ID if not passed by URL or query parameter.
        :arg scroll: Specify how long a consistent view of the index should be
            maintained for scrolled search
        """
        if scroll_id in NULL_VALUES and body in NULL_VALUES:
            raise ValueError("You need to supply scroll_id or body.")
        elif scroll_id and not body:
            body = scroll_id
        elif scroll_id:
            query_params[EsConst.SCROLL_ID] = scroll_id

        path = self._es_parser.make_path(EsMethods.SEARCH, EsMethods.SCROLL)
        result = yield self._perform_request(HttpMethod.GET, path, body, params=query_params)
        returnValue(result)

    @inlineCallbacks
    def clear_scroll(self, scroll_id=None, body=None, **query_params):
        """
        Clear the scroll request created by specifying the scroll parameter to
        search.
        `<http://www.elastic.co/guide/en/elasticsearch/reference/current/search-request-scroll.html>`_
        :param scroll_id: A comma-separated list of scroll IDs to clear
        :param body: A comma-separated list of scroll IDs to clear if none was
            specified via the scroll_id parameter
        """
        if scroll_id in NULL_VALUES and body in NULL_VALUES:
            raise ValueError("You need to supply scroll_id or body.")
        elif scroll_id and not body:
            body = scroll_id
        elif scroll_id:
            query_params[EsConst.SCROLL_ID] = scroll_id

        path = self._es_parser.make_path(EsMethods.SEARCH, EsMethods.SCROLL)
        result = yield self._perform_request(HttpMethod.DELETE, path, body, params=query_params)
        returnValue(result)

    @inlineCallbacks
    def scan(self, index, doc_type, query=None, scroll='5m', preserve_order=False, size=10, **kwargs):
        """
        Simple abstraction on top of the
        :meth:`~elasticsearch.Elasticsearch.scroll` api - a simple iterator that
        yields all hits as returned by underlining scroll requests.

        By default scan does not return results in any pre-determined order. To
        have a standard order in the returned documents (either by score or
        explicit sort definition) when scrolling, use ``preserve_order=True``. This
        may be an expensive operation and will negate the performance benefits of
        using ``scan``.
        :param index: the index to query on
        :param doc_type: the doc_type to query on
        :param query: body for the :meth:`~elasticsearch.Elasticsearch.search` api
        :param scroll: Specify how long a consistent view of the index should be
            maintained for scrolled search
        :param preserve_order: don't set the ``search_type`` to ``scan`` - this will
            cause the scroll to paginate with preserving the order. Note that this
            can be an extremely expensive operation and can easily lead to
            unpredictable results, use with caution.
        :param size: the number of results to fetch in each scroll query

        Any additional keyword arguments will be passed to the initial
        :meth:`~elasticsearch.Elasticsearch.search` call::

            scan(index="coding_languages",
                doc_type="languages_description",
                query={"query": {"match": {"title": "python"}}},
                index="orders-*",
                doc_type="books"
            )

        """
        if not preserve_order:
            kwargs['search_type'] = 'scan'
        # initial search
        results = yield self.search(index=index,
                                    doc_type=doc_type,
                                    body=query,
                                    size=size,
                                    scroll=scroll,
                                    **kwargs)

        returnValue(Scroller(self, results, scroll, size))

    @inlineCallbacks
    def count(self, index=None, doc_type=None, body=None, **query_params):
        """
        Execute a query and get the number of matches for that query.
        `<http://www.elastic.co/guide/en/elasticsearch/reference/current/search-count.html>`_
        :param index: A comma-separated list of indices to restrict the results
        :param doc_type: A comma-separated list of types to restrict the results
        :param body: A query to restrict the results specified with the Query DSL
            (optional)
        :arg allow_no_indices: Whether to ignore if a wildcard indices
            expression resolves into no concrete indices. (This includes `_all`
            string or when no indices have been specified)
        :arg analyze_wildcard: Specify whether wildcard and prefix queries
            should be analyzed (default: false)
        :arg analyzer: The analyzer to use for the query string
        :arg default_operator: The default operator for query string query (AND
            or OR), default 'OR', valid choices are: 'AND', 'OR'
        :arg df: The field to use as default where no field prefix is given in
            the query string
        :arg expand_wildcards: Whether to expand wildcard expression to concrete
            indices that are open, closed or both., default 'open', valid
            choices are: 'open', 'closed', 'none', 'all'
        :arg ignore_unavailable: Whether specified concrete indices should be
            ignored when unavailable (missing or closed)
        :arg lenient: Specify whether format-based query failures (such as
            providing text to a numeric field) should be ignored
        :arg lowercase_expanded_terms: Specify whether query terms should be
            lowercased
        :arg min_score: Include only documents with a specific `_score` value in
            the result
        :arg preference: Specify the node or shard the operation should be
            performed on (default: random)
        :arg q: Query in the Lucene query string syntax
        :arg routing: Specific routing value
        """
        if doc_type and not index:
            index = EsConst.ALL_VALUES

        path = self._es_parser.make_path(index, doc_type, EsMethods.COUNT)
        result = yield self._perform_request(HttpMethod.GET, path, body, params=query_params)
        returnValue(result)

    @inlineCallbacks
    def bulk(self, body, index=None, doc_type=None, **query_params):
        """
        Perform many index/delete operations in a single API call.
        See the :func:`~elasticsearch.helpers.bulk` helper function for a more
        friendly API.
        `<http://www.elastic.co/guide/en/elasticsearch/reference/current/docs-bulk.html>`_
        :param body: The operation definition and data (action-data pairs),
            separated by newlines
        :param index: Default index for items which don't provide one
        :param doc_type: Default document type for items which don't provide one
        :arg consistency: Explicit write consistency setting for the operation,
            valid choices are: 'one', 'quorum', 'all'
        :arg fields: Default comma-separated list of fields to return in the
            response for updates
        :arg pipeline: The pipeline id to preprocess incoming documents with
        :arg refresh: Refresh the index after performing the operation
        :arg routing: Specific routing value
        :arg timeout: Explicit operation timeout
        """
        self._es_parser.is_not_empty_params(body)
        path = self._es_parser.make_path(index, doc_type, EsMethods.BULK)
        result = yield self._perform_request(HttpMethod.POST,
                                             path,
                                             self._bulk_body(body),
                                             params=query_params)
        returnValue(result)

    @inlineCallbacks
    def msearch(self, body, index=None, doc_type=None, **query_params):
        """
        Execute several search requests within the same API.
        `<http://www.elastic.co/guide/en/elasticsearch/reference/current/search-multi-search.html>`_
        :arg body: The request definitions (metadata-search request definition
            pairs), separated by newlines
        :arg index: A comma-separated list of index names to use as default
        :arg doc_type: A comma-separated list of document types to use as
            default
        :arg search_type: Search operation type, valid choices are:
            'query_then_fetch', 'query_and_fetch', 'dfs_query_then_fetch',
            'dfs_query_and_fetch'
        """
        self._es_parser.is_not_empty_params(body)
        path = self._es_parser.make_path(index,
                                         doc_type,
                                         EsMethods.MULTIPLE_SEARCH)

        result = yield self._perform_request(HttpMethod.GET,
                                             path,
                                             self._bulk_body(body),
                                             params=query_params)
        returnValue(result)

    @inlineCallbacks
    def _perform_request(self, method, path, body=None, params=None):
        url = self._es_parser.prepare_url(self._hostname, path, params)

        if body is not None and not isinstance(body, string_types):
            body = json.dumps(body)
        try:
            response = yield self._async_http_client.request(method,
                                                             url,
                                                             data=body,
                                                             timeout=self._timeout,
                                                             auth=self._auth)

            content = yield self._get_content(response)

            if response.code in (ResponseCodes.OK,
                                 ResponseCodes.CREATED,
                                 ResponseCodes.ACCEPTED):
                returnValue(content)

            if response.code == ResponseCodes.NOT_FOUND:
                raise NotFoundError(content)

            if response.code == ResponseCodes.BAD_REQUEST:
                raise RequestError(content)

            # This is a place holder for unknown exceptions
            # that haven't been encaplulated yet
            msg_fmt = "unknown error; code: {code} | message: {msg}"
            raise ElasticsearchException(msg_fmt.format(code=response.code,
                                                        msg=str(content)))

        except (ResponseNeverReceived,
                CancelledError,
                ConnectingCancelledError) as e:
            raise ConnectionTimeout(str(e))

    @inlineCallbacks
    def _get_content(self, response):
        content = None
        try:
            content = yield response.json()
        except ValueError as e:
            content_txt = yield response.content()
            content = json.loads(content_txt)
        except Exception as e:
            # unknown exceptions are ignored
            # and the content is set to None
            pass
        finally:
            returnValue(content)

    @staticmethod
    def _bulk_body(body):
        # if not passed in a string, serialize items and join by newline
        line_feed = '\n'
        if not isinstance(body, str):
            body = line_feed.join(map(json.dumps, body))

        # bulk body must end with a newline
        if not body.endswith(line_feed):
            body += line_feed

        return body



import sys

PY2 = sys.version_info.major == 2

if PY2:
    string_types = basestring,
    from urllib import quote, urlencode
    from urlparse import  urlparse
    from itertools import imap as map
else:
    string_types = str, bytes
    from urllib.parse import quote, urlencode, urlparse
    map = map


#!/usr/bin/env python

from setuptools import setup, find_packages
from pip.req import parse_requirements
from distutils.util import convert_path
import codecs
import sys

# Load version information
main_ns = {}
ver_path = convert_path('twistes/version.py')
with codecs.open(ver_path, 'rb', 'utf8') as ver_file:
    exec (ver_file.read(), main_ns)

install_requires = ['twisted==16.2.0', 'treq==15.1.0']

if sys.version_info < (2, 7):
    # python 2.6 isn't supported
    raise RuntimeError('This version requires Python 2.7+')

setup(
    name='twistes',
    version=main_ns['__version__'],
    author='Avihad Menahem',
    author_email='avihad87@gmail.com',
    description='Asynchronous elastic search client using Twisted and Treq',
    license='Apache License Version 2.0',
    platforms=['Any'],
    keywords=[
        'elasticsearch', 'async', 'asynchronous', 'twisted', 'treq'
    ],
    url='https://github.com/avihad/twistes',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Testing'
    ],
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    download_url='https://github.com/avihad/twistes/tarball/{version}'.format(version=main_ns['__version__']),
    zip_safe=False,
    include_package_data=True,
    install_requires=install_requires,
    test_suite='tests'
)



from twistes.consts import EsConst
from twistes.exceptions import ScanError


class EsUtils(object):

    @staticmethod
    def extract_hits(results):
        EsUtils.validate_scan_result(results)
        if EsUtils.has_results(results):
            return results[EsConst.HITS][EsConst.HITS]
        else:
            return []

    @staticmethod
    def has_results(results):
        return results and \
               EsConst.HITS in results and \
               EsConst.HITS in results[EsConst.HITS] and \
               EsConst.TOTAL in results[EsConst.HITS] and \
               results[EsConst.HITS][EsConst.TOTAL] > 0 and \
               results[EsConst.HITS][EsConst.HITS]

    @staticmethod
    def is_get_query_with_results(results):
        """
        :param results: the response from Elasticsearch
        :return: true if the get query returned a result, false otherwise
        """
        return results and EsConst.FOUND in results and results[EsConst.FOUND] and EsConst.FIELDS in results

    @staticmethod
    def validate_scan_result(results):
        """ Check if there's a failed shard in the scan query"""
        if results[EsConst.SHARDS][EsConst.FAILED] and results[EsConst.SHARDS][EsConst.FAILED] > 0:
            raise ScanError(
                'Scroll request has failed on %d shards out of %d.' %
                (results[EsConst.SHARDS][EsConst.FAILED], results[EsConst.SHARDS][EsConst.TOTAL])
            )
